package nom.paolo.ai.game.objects;

import java.awt.Color;
import java.awt.Image;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.Stack;

import nom.paolo.ai.game.animation.Sprite;
import nom.paolo.ai.game.command.AgentCommand;
import nom.paolo.ai.game.command.MoveAgentCommand;
import nom.paolo.ai.game.command.StayPutCommand;
import nom.paolo.ai.game.display.Displayable;
import nom.paolo.ai.game.display.Viewport;
import nom.paolo.ai.game.map.GameMap;
import nom.paolo.ai.util.ExperimentStats;
import nom.paolo.ai.util.MathUtils;

/**
 * 
 * @author Paolo Victor, paolo@dsc.ufcg.edu.br
 */
public class Agent implements Runnable, Anchor, Displayable {
	/**
	 * The speed on which an agent may move
	 */
	private static final double MAX_AGENT_SPEED = 0.75;

	/**
	 * Key for this agent's "Dead" animation
	 */
	public static final String DEAD_ANIMATION = "dead";
	
	/**
	 * Key for this agent's "Idle" animation
	 */
	public static final String IDLE_ANIMATION = "idle";
	
	/**
	 * This agent's field of view, in degrees
	 */
	public static final double FIELD_OF_VIEW = 90;
	
	/**
	 * The maximum distance that this agent's sight can reach
	 */
	public static final double MAX_SIGHT_DISTANCE = 150;
	
	/**
	 * The maximum amount of time that this agent will keep
	 * locked to a target even if it can't see it
	 */
	public static final int MAX_ENEMY_SIGHT_LOCK = 100;
	
	/**
	 * The maximum number of members in a team
	 */
	public int MAX_SQUAD_SIZE;
	
	/**
	 * Identifier for the red team
	 */
	public static final int TEAM_RED = 0;
	
	/**
	 * Identifier for the blue team
	 */
	public static final int TEAM_BLUE = 1;
	
	/**
	 * Identifier for the green team
	 */
	public static final int TEAM_GREEN = 2;
	
	/**
	 * Identifier for the yellow team
	 */
	public static final int TEAM_YELLOW = 3;

	/**
	 * The initial energy for an agent
	 */
	public static final double INITIAL_ENERGY = 50;

	/**
	 * The maximum amount of time that this agent has
	 * to wait before attacking
	 */
	public static final int ATTACK_DELAY = 20;

	/**
	 * The maximum damage that an attack can deal
	 */
	private static final double MAXIMUM_DAMAGE = 8;

	/**
	 * The amount of time before the agents checks its anchor's
	 * location again
	 */
	private static final int ANCHOR_REVALIDATION_TIME = 75;
	
	/**
	 * This agent's id
	 */
	protected String id;
	
	/**
	 * This agent team's identifier
	 */
	protected int teamId;
	
	/**
	 * The time since this agent hasn't seen an enemy 
	 */
	protected int lastEnemySight = 0;			
	
	/**
	 * This agent's action stack
	 */
	protected Stack<AgentCommand> actions;
	
	/**
	 * The order that has to be followed by this agent
	 */		
	protected AgentCommand order = null;	
	
	/**
	 * The action that is currently being executed by this agent
	 */
	protected AgentCommand actualAction = null;
	
	/**
	 * This agent's anchor
	 */
	protected Anchor anchor = null;
	
	/**
	 * This agent's target
	 */
	protected Agent target = null;	
	
	/**
	 * This agent's energy
	 */
	protected double energy;
	
	/**
	 * The amount of time that this agent will have to
	 * wait for its next attack
	 */
	protected int attackDelay;
	
	/**
	 * This agent's graphical avatar
	 */
	protected Sprite sprite;
	
	/**
	 * The squad of which this agent is part
	 */
	protected Set<Agent> squad;
	
	/**
	 * The other agents on the environment
	 */
	protected List<Agent> others;
	
	/**
	 * This agent's X position
	 */
	protected double x;
	
	/**
	 * This agent's Y position
	 */
	protected double y;
	
	/**
	 * This agent's horizontal speed
	 */
	protected double xSpeed;
	
	/**
	 * This agent's vertical speed
	 */
	protected double ySpeed;
	
	/**
	 * The angle that this agent is looking at, in radians
	 */
	protected double viewAngle;
	
	/**
	 * The last X coordinate where the agent saw its target
	 */
	protected double lastEnemySightX;
	
	/**
	 * The last X coordinate where the agent saw its target
	 */
	protected double lastEnemySightY;
	
	/**
	 * The amount of time before the agents checks its anchor's
	 * location again
	 */
	protected int revalidateAnchorLocation = ANCHOR_REVALIDATION_TIME;
	
	/**
	 * Default constructor
	 * 
	 * @param teamId The agent team's identifier
	 * @param x The initial X position
	 * @param y The initial Y position
	 * @param agentEnergy
	 * @param maxSquadSize
	 * @param others The other agents on the environment
	 */
	public Agent( String id, int teamId, int agentEnergy, int maxSquadSize, double x, double y, List<Agent> others ) {
		this.actions = new Stack<AgentCommand>();
		
		this.id = id;
		this.teamId = teamId;
		
		this.x = x;
		this.y = y;
		
		this.others = others;
		
		this.xSpeed = 0;
		this.ySpeed = 0;
		
		this.energy = agentEnergy;
		this.attackDelay = ATTACK_DELAY;
		
		this.MAX_SQUAD_SIZE = maxSquadSize;
		
		// The initial view angle is random
		this.viewAngle = Math.random() * 2 * Math.PI;
		
		// Initial squad is an one-man-army =)
		this.squad = new HashSet<Agent>();
		this.squad.add( this );
	}
	
	public void run() {
		while( true ) {
			// Check if the agent is dead. If it's dead,
			// set animation to death, stop moving, warn the squad
			// and stop thread
			if( ! this.isAlive() ) {
				this.stopMoving();
				this.announceDeath();
				this.sprite.setAnimation( DEAD_ANIMATION );
					
				return;
			}
			
			// Move the agent around...
			this.updatePosition();
			
			// Updating view angle
			this.updateViewAngle();			
			
			// If there are orders to follow, follow'em!
			this.followOrders();
			
			this.updateAnchorLocation();
			
			// If there are no orders to follow or no actual action,
			// execute the next action or execute an idle action
			this.executeActions();
			
			// Scanning other agents
			this.scanAgents();
			
			// Dealing with the target
			this.dealWithTarget();			
			
			// Sleep for a while
			try {
				Thread.sleep( 100/6 );
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
	}
	
	/**
	 * Updates the agent's movement, setting it actual 
	 * action to none. On the next update, it'll move
	 * near the anchor.
	 */
	private void updateAnchorLocation() {
		if( target != null || anchor != null ) {
			revalidateAnchorLocation--;
			if( revalidateAnchorLocation <= 0 ) {
				revalidateAnchorLocation = 100;
				
				if( this.actualAction != null ) {
					this.actualAction.stop();
					this.stopMoving();
				}
			}
		}
	}

	/**
	 * Updates the agent's view angle. If it has a target, look at it,
	 * if it hasn't, look at the walking direction.
	 *
	 */
	private void updateViewAngle() {
		if( target != null ) {				
			this.viewAngle = Math.atan2( target.getY() - this.getY(), target.getX() - this.getX() );
		} else {
			double targetAngle = viewAngle;

			if( xSpeed != 0.0 || ySpeed != 0.0 ) {
				targetAngle = Math.atan2( ySpeed, xSpeed );
			}
				
			if( MathUtils.angleBetween( targetAngle, viewAngle ) > 0.1 ) { 
				double viewAngleDec = viewAngle - 0.1;
				double viewAngleInc = viewAngle + 0.1;
				
				if( MathUtils.angleBetween( targetAngle, viewAngleDec ) <
					MathUtils.angleBetween( targetAngle, viewAngleInc ) ) {
					viewAngle = viewAngleDec;
				} else {
					viewAngle = viewAngleInc;
				}
			}
		}
	}

	/**
	 * Takes the necessary actions to pursue/attack the target, if there's one.
	 */
	private void dealWithTarget() {
		// Updates the weapon's cool down timer
		this.attackDelay--;
		
		// If the agent has a target...		
		if( target != null ) {
			// If the target is dead, forget it
			if( ! target.isAlive() ) {
				this.resetTarget();				
				return;
			}			
			
			// If the agent can't see the target, increase
			// the "I can't see it" counter
			if( ! this.canSee( target ) ) {
				this.lastEnemySight++;
			} else {						
				// If it can see it, reset the "I can't see it" counter,
				// record its last seen position and attack it
				this.lastEnemySight = 0;			
				
				this.lastEnemySightX = target.getX();
				this.lastEnemySightY = target.getY();			
					
				// If the weapon has cooled down, attack
				if( this.attackDelay <= 0 ) {
					// Setting the weapon's heat to maximum
					this.attackDelay = ATTACK_DELAY;
					
					// Attacks the target
					this.attackAgent( target );
				}					
			}
				
			// If enemy is out of sight for too much time, let it go
			if( this.lastEnemySight > MAX_ENEMY_SIGHT_LOCK ) {
				// Reset the "I can't see it" counter
				this.lastEnemySight = 0;
				
				// Reset the target
				this.resetTarget();	
				
				// Move to the last location that the target seen
				this.actions.push( new MoveAgentCommand( this, (int)lastEnemySightX, (int)lastEnemySightY, (float)MAX_AGENT_SPEED ) );
			}
		}
	}
	
	/**
	 * Makes this agent "forget" its current target
	 * and stop any actions that it would be executing
	 * to pursue/attack it.
	 */
	private void resetTarget() {
		// Set target as "none"
		this.target = null;
		
		// Stop whatever action it was executing
		if( this.actualAction != null ) {
			this.actualAction.stop();
		}
	}

	/**
	 * Executes the agent's actual action, or executes the
	 * next action if the actual action is finished.
	 */
	private void executeActions() {		
		if( this.isIdle() || this.actualAction.isFinished() ) {
			if( this.hasMoreActions() ) {
				this.executeNextAction();
			} else {
				this.takeAction();
			}
		} else {
			this.actualAction.execute();
		}		
	}

	/**
	 * If there's an order to follow, execute it.
	 */
	private void followOrders() {
		if( order != AgentCommand.NO_COMMAND ) {
			// If the character is not idle, push the action
			// on the actions stack to be executed later
			if( ! this.isIdle() ) {
				this.actions.push( actualAction );
			}
			
			this.actualAction = order;
			order = AgentCommand.NO_COMMAND;
		}
	}

	/**
	 * Updates the agent's X and Y coordinates, according to its speed
	 */
	private void updatePosition() {
		this.x += xSpeed;
		this.y += ySpeed;
	}

	/**
	 * Attacks another agent. The damage formula is:
	 * 
	 * MAX_SIGHT_DISTANCE / distance from tbe agent to the target 
	 * 
	 * @param target The target that will be attacked
	 */
	private void attackAgent( Agent target ) {		
		double damage = ( MAX_SIGHT_DISTANCE - this.distanceBetween( target ) ) / MAX_SIGHT_DISTANCE * MAXIMUM_DAMAGE + Math.random() * MAXIMUM_DAMAGE / 3;
		
		this.energy += damage / 2;
		
		if( damage > 0.0 ) {		
			target.hurt( this, damage );
		}
	}

	/**
	 * Hurts this agent, subtracting a given amount from its anergy.
	 * If the hurt agent hasn't a target of the offender is nearer
	 * than the current target, the offender will be the new target,
	 * 
	 * @param offender The agent that hurt this agent
	 * @param damage
	 */
	public void hurt( Agent offender, double damage ) {
		this.energy -= damage;
		
		if( this.target == null || this.distanceBetween( this.target ) > this.distanceBetween( offender ) ) {
			this.setAsTarget( offender );
		}
	}

	/**
	 * Iterates through the environment's agents and 
	 * executes the necessary actions (such as set as target,
	 * join squad, etc.
	 * 
	 * Algoritm:
	 * 
	 * For each of the environment's targets:
	 * 
	 * - If the agent is a foe and there's no target, set 
	 *   it as the new target
	 * - If the agent is a friend and I don't have a squad,
	 *   try to join its squad
	 * - If the agent is on my squad and I don't have a target,
	 *   attack its target
	 */
	private void scanAgents() {
		for( Agent other : others ) {
			// If the other agent isn't myself ant it's alive...
			if( ! other.getId().equals( this.id ) && other.isAlive() ) {
				
				// If it's a foe
				if( other.getTeamId() != this.teamId ) {
					// If I can see the foe ant I still don't have a target
					if( this.target == null && this.canSee( other ) ) {
						this.setAsTarget( other );
					}
					
				} else {
					// If i'm alone, try to join a squad
					if( this.squad.size() == 1 ) {
						if( other.requestSquad( this ) ) {
							this.squad.addAll( other.getSquad() );
							
							if( other.anchor != null ) {
								this.anchor = other.anchor;
							} else {
								this.anchor = other;
							}
						}
					} else { // I'm on a team!
						// Attack team mate's target
						if( squad.contains( other ) ) {
							if( other.target != null && this.target == null ) {
								ExperimentStats.getInstance().addTargetFriendLock();
								this.setAsTarget( other.target );
							}
						}
					}
				}
			}
		}
	}
	
	/**
	 * Sets an agent as this agent's target
	 * @param other
	 */
	private void setAsTarget(Agent other) {
		ExperimentStats.getInstance().addTargetLock();
		this.target = other;
		this.lastEnemySight = 0;
		this.actions.push( this.actualAction );		
		this.actualAction = AgentCommand.NO_COMMAND;
	}

	/**
	 * Ask this agent to enter its squad. If the request is accepted,
	 * 
	 * 
	 * @param requester
	 * @return true if the request was accepted
	 */
	private synchronized boolean requestSquad( Agent requester ) {
		// I'm the leader and the team is full. Refuse.
		if( this.getSquad().size() >= MAX_SQUAD_SIZE && anchor == null ) {
			return false;
		}
		
		// Hey, there's room in my team!
		if( this.getSquad().size() < MAX_SQUAD_SIZE ) {
			this.reportInSquad( requester );			
			for( Agent mate : squad ) {
				mate.reportInSquad( requester );
			}			
			
			return true;
		}
		
		return false;
	}	
	
	/**
	 * Adds the given agent on this agent's squad
	 * @param agent
	 */
	private synchronized void reportInSquad( Agent agent ) {
		this.squad.add( agent );
	}

	/**
	 * Removes the given agent from this agent's squad. If
	 * the given agent is this agent's anchor, it resets the
	 * squad.
	 * 
	 * @param agent
	 */
	private synchronized void reportOutOfSquad( Agent agent ) {
		synchronized( squad ) {
			this.squad.remove( agent );
			
			if( agent == this.anchor ) {
				this.anchor = null;
				
				this.squad = new HashSet<Agent>();
				this.squad.add( this );
			}
		}
	}	

	/**
	 * Tells this agent's squad that it is dead.
	 */
	private void announceDeath() {
		synchronized( squad ) {
			for( Agent mate : squad ) {
				if( !mate.getId().equals( this.id ) ) {						
					mate.reportOutOfSquad( this );
				}
			}
		}
	}

	/**
	 * @return This agent's squad
	 */
	public Set<Agent> getSquad() {
		return this.squad;
	}
	
	/**
	 * @param other
	 * @return true if the other agent is on the line of sight
	 */
	public boolean canSee( Agent other ) {
		if( this.distanceBetween( other ) > MAX_SIGHT_DISTANCE ) {
			return false;
		}
//		
//		double directAngle = Math.atan2( other.getY() - this.getY(), other.getX() - this.getX() );
//		double viewAngleA = directAngle - Math.toRadians( FIELD_OF_VIEW / 2 );
//		double viewAngleB = directAngle + Math.toRadians( FIELD_OF_VIEW / 2 );
//		
//		return this.viewAngle >= viewAngleA && this.viewAngle <= viewAngleB &&
//			   GameMap.getInstance().lineMayBeDraw( this.getX(), this.getY(), other.getX(), other.getY(), this.getTeamId() );
		
		
		double directAngle = Math.atan2( other.getY() - this.getY(), other.getX() - this.getX() );
//		double viewAngleA = directAngle - Math.toRadians( FIELD_OF_VIEW / 2 );
//		double viewAngleB = directAngle + Math.toRadians( FIELD_OF_VIEW / 2 );
		
		return MathUtils.angleBetween( viewAngle, directAngle ) <= FIELD_OF_VIEW / 2 &&
			   GameMap.getInstance().lineMayBeDraw( this.getX(), this.getY(), other.getX(), other.getY(), this.getTeamId() );
	}

	/**
	 * Selects the next action from this agent's action stack
	 */
	private void executeNextAction() {
		this.actualAction = this.hasMoreActions() ? actions.pop() : AgentCommand.NO_COMMAND;
	}

	/**
	 * @return true if there are more actions to be executed by this agent
	 */
	private boolean hasMoreActions() {
		return !actions.isEmpty();
	}

	/**
	 * @return true if the agent isn't executing any action
	 */
	private boolean isIdle() {
		return actualAction == AgentCommand.NO_COMMAND;
	}

	/**
	 * Selects an action based on the agent's target and anchor.
	 * 
	 * 1. If there's a target, the agent will walk around it.
	 * 2. If there's an anchor, the agent will walk around it.
	 * 3. If there's neither target nor anchor, just walk around 
	 */
	private void takeAction() {
		if( target != null ) {
			this.walkAround( target, (int)MAX_SIGHT_DISTANCE / 3, (int)MAX_SIGHT_DISTANCE / 2 );
		} else if( anchor != null ) {
			this.walkAround( anchor, 0, (int)MAX_SIGHT_DISTANCE / 10 );
		} else {
			this.walkAround();
			this.actions.push( new StayPutCommand( this, (int)(Math.random() * 10 + 10) ) );
		}
	}

	/**
	 * Selects a random point on the map and adds a "walk to" action
	 * to make the agent walk towards it
	 */
	private void walkAround() {
		int[][] map = GameMap.getInstance().getVisibleMap( this.teamId );
		int tileSize = GameMap.getInstance().getTileSize();
					
		int xD = (int)(Math.random() * (map[0].length - 1) );
		int yD = (int)(Math.random() * (map.length - 1) );
		
		while( map[yD][xD] > 0 ) {
			xD = (int)(Math.random() * (map[0].length - 1) );
			yD = (int)(Math.random() * (map.length - 1) );
		}
		
		int xTarget = xD * tileSize;
		int yTarget = yD * tileSize;
	
		this.actions.push( new MoveAgentCommand( this, xTarget, yTarget, MAX_AGENT_SPEED ) );		
	}
	
	/**
	 * Selects a random point on the map that follows the conditions:
	 * 
	 * 1. The anchor must be able to be seen from it
	 * 2. It must be at a least at minimumDistance from the anchor
	 * 3. It must be at a least at maximumDistance from the anchor
	 * 
	 * and adds a "walk to" action to make the agent walk towards it.
	 */
	private void walkAround( Anchor a, int minimumDistance, int maximumDistance ) {
		int[][] map = GameMap.getInstance().getVisibleMap( this.teamId );
		int tilesize = GameMap.getInstance().getTileSize();
					
		int nextX = (int)(Math.random() * (map[0].length - 1) * tilesize );
		int nextY = (int)(Math.random() * (map.length - 1) * tilesize );
		
		while( ! mayMoveTo( a, map, tilesize, nextX, nextY, minimumDistance, maximumDistance ) ) {
			nextX = (int)(Math.random() * (map[0].length - 1) * tilesize );
			nextY = (int)(Math.random() * (map.length - 1) * tilesize );
		}		
				
		this.actualAction = new MoveAgentCommand( this, nextX, nextY, MAX_AGENT_SPEED );
	}
	
	/**
	 * Tells if the point satisfies the conditions:
	 * 
	 * 1. The anchor must be able to be seen from it
	 * 2. It must be at a least at minimumDistance from the anchor
	 * 3. It must be at a least at maximumDistance from the anchor
	 * 
	 * @return 
	 */
	private boolean mayMoveTo( Anchor a, int[][] map, int tilesize, int px, int py, int minimumDistance, int maximumDistance ) {
		double distanceToPoint = MathUtils.distanceBetween( px, py, a.getX(), a.getY() );
				
		return map[ py/tilesize ][ px/tilesize ] == 0
		     && distanceToPoint > minimumDistance
		     && distanceToPoint < maximumDistance
		     && GameMap.getInstance().lineMayBeDraw( px, py, a.getX(), a.getY(), this.getTeamId() );
	}
	
	/**
	 * Gives this agent an order
	 * @param order
	 */
	public void giveOrder( AgentCommand order ) {
		this.order = order;
	}
	
	/**
	 * @return Wether this agent is alive or not ( an agent
	 * is alive if its energy is greater than 0)
	 */
	public boolean isAlive() {
		return this.energy > 0.0;
	}
	

	/**
	 * @return This agent's energy
	 */
	public double getEnergy() {
		return this.energy;
	}

	///////////////////////////////////////////////////////////
	// Anchor methods                                        //
	///////////////////////////////////////////////////////////
	public int getX() {
		return (int)this.x;
	}

	public int getY() {
		return (int)this.y;
	}
	
	public double distanceBetween(Anchor anchor) {
		if( anchor == null ) {
			return 0;
		}
		
		double dx2 = Math.pow(this.getX() - anchor.getX(), 2);
		double dy2 = Math.pow(this.getY() - anchor.getY(), 2);
		
		return Math.sqrt( dx2 + dy2 );
	}

	///////////////////////////////////////////////////////////
	// Sprite methods                                        //
	///////////////////////////////////////////////////////////
	public void setSprite(Sprite sprite) {
		this.sprite = sprite;
	}
	
	public Image getImage() {
		return this.sprite.getImage();
	}

	public int getDisplayPositionX() {
		return (int)this.x - this.getWidth() / 2;
	}

	public int getDisplayPositionY() {
		return (int)this.y - this.getHeight() / 2;
	}

	public double getRotateAngle() {
		return this.viewAngle;
	}

	public int getWidth() {
		return this.sprite.getWidth();
	}

	public int getHeight() {
		return this.sprite.getHeight();
	}
	
	public void updateSprite() {
		this.sprite.updateObject();
	}

	public int getTeamId() {
		return this.teamId;
	}
	
	public String getId() {
		return this.id;
	}
	
	public void displayInfo( Viewport viewport ) {
		if( this.isAlive() ) {
			// Drawing laser line
			if( this.target != null ) {
				synchronized( target ) {
					if( this.attackDelay > ATTACK_DELAY / 2 ) {
						double colorFactor = ((attackDelay - ATTACK_DELAY / 2.0) / (ATTACK_DELAY / 2.0) );
						Color color = new Color( 1, 0, 0, (float)colorFactor );
						viewport.drawLine( this.getX(), this.getY(), target.getX(), target.getY(), color );	
					}
				}
			}
		}
	}

	///////////////////////////////////////////////////////////
	// Movement methods                                      //
	///////////////////////////////////////////////////////////
	/**
	 * Tells the agent to move to the right
	 * 
	 * @param speed the speed
	 */
	public void moveRight( double speed ) {
		this.xSpeed = speed;
	}
	
	/**
	 * Tells the agent to move to the right
	 * 
	 * @param speed the speed
	 */
	public void moveLeft( double speed ) {
		this.xSpeed = -speed;
	}
	
	/**
	 * Tells the agent to move down
	 * 
	 * @param speed the speed
	 */
	public void moveDown( double speed ) {
		this.ySpeed = speed;
	}
	
	/**
	 * Tells the agent to move up
	 * 
	 * @param speed the speed
	 */
	public void moveUp( double speed ) {
		this.ySpeed = -speed;
	}

	/**
	 * Tells the agent to stop moving
	 */
	public void stopMoving() {
		this.xSpeed = 0;
		this.ySpeed = 0;
	}
	
	public String toString() {
		NumberFormat formatter = new DecimalFormat( "0.0" );
		return id + " (" + formatter.format( this.energy ) +") - Target: " + (target != null ? target.getId() : "No target");
	}
	
	public int hashCode() {
		return this.id.hashCode();
	}

	public int compareTo(Object o) {
		if( ! ( o instanceof Agent ) ) {
			return -1;
		}
		
		if( this.isAlive() ) {
			return 1;
		} else  {
			return -1;
		}
	}
}
