package nom.paolo.ai.game;

import java.awt.Color;
import java.awt.Image;
import java.awt.geom.Point2D;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.LinkedList;
import java.util.List;
import java.util.Scanner;
import java.util.StringTokenizer;

import javax.imageio.ImageIO;
import javax.swing.JPanel;

import nom.paolo.ai.game.animation.Animation;
import nom.paolo.ai.game.animation.BaseSprite;
import nom.paolo.ai.game.animation.Sprite;
import nom.paolo.ai.game.display.Displayable;
import nom.paolo.ai.game.display.Viewport2D;
import nom.paolo.ai.game.map.GameMap;
import nom.paolo.ai.game.objects.Agent;

public class GameCanvas extends JPanel {
	private List<Agent> agents;
	private List<Displayable> objects;
	private Viewport2D viewport;
	private int lockedAgent; 
	
	public GameCanvas( String defs, int width, int height, Image buffer ) throws FileNotFoundException {
		this.objects = new LinkedList<Displayable>();		
		this.agents = new LinkedList<Agent>();
		this.viewport = new Viewport2D( buffer.getGraphics(), width, height );
		
		this.lockedAgent = 0;
		
		this.load( defs );		
	}
	
	public void addAgent( String id, int teamId, int agentEnergy, int maxSquadSize ) throws IOException {
		Point2D spawnPoint = GameMap.getInstance().getSpawnPoint( teamId );
		
		Agent agent = new Agent( id, teamId, agentEnergy, maxSquadSize, spawnPoint.getX(), spawnPoint.getY(), agents );
		
		agents.add( agent );
		objects.add( agent );
		
		Sprite sprite = new BaseSprite( 100, 100, 16, 16, 0.0 );
		if( teamId == Agent.TEAM_BLUE ) {
			sprite.attachAnimation( "idle", new Animation( ImageIO.read( new File( "data/images/agent_blue.png" ) ), 32, 32, 0, 0, 1, 100 ) );
			sprite.attachAnimation( "dead", new Animation( ImageIO.read( new File( "data/images/agent_blue_dead.png" ) ), 32, 32, 0, 0, 1, 100 ) );			
		} else if ( teamId == Agent.TEAM_RED ) {
			sprite.attachAnimation( "idle", new Animation( ImageIO.read( new File( "data/images/agent_red.png" ) ), 32, 32, 0, 0, 1, 100 ) );
			sprite.attachAnimation( "dead", new Animation( ImageIO.read( new File( "data/images/agent_red_dead.png" ) ), 32, 32, 0, 0, 1, 100 ) );
		} else if ( teamId == Agent.TEAM_GREEN ) {
			sprite.attachAnimation( "idle", new Animation( ImageIO.read( new File( "data/images/agent_green.png" ) ), 32, 32, 0, 0, 1, 100 ) );
			sprite.attachAnimation( "dead", new Animation( ImageIO.read( new File( "data/images/agent_green_dead.png" ) ), 32, 32, 0, 0, 1, 100 ) );
		} else if ( teamId == Agent.TEAM_YELLOW ) {
			sprite.attachAnimation( "idle", new Animation( ImageIO.read( new File( "data/images/agent_yellow.png" ) ), 32, 32, 0, 0, 1, 100 ) );
			sprite.attachAnimation( "dead", new Animation( ImageIO.read( new File( "data/images/agent_yellow_dead.png" ) ), 32, 32, 0, 0, 1, 100 ) );
		}
		sprite.setAnimation( "idle" );
		agent.setSprite( sprite );
	}
	
	public void cycleCamera() {		
		if( agents.size() > 0 ) {
			this.lockedAgent = (this.lockedAgent + 1 ) % agents.size();
			
			while( ! agents.get( lockedAgent ).isAlive() ) {
				this.lockedAgent = (this.lockedAgent + 1 ) % agents.size();	
			}
		}
	}
	
	public void display() {
		if( agents.size() > 0 ) {
			this.viewport.lockOn( agents.get( lockedAgent ) );
			
		}
		
		viewport.drawObjects( objects );
		
		for( Agent agent : agents ) {
			if( agent.isAlive() ) {
				int energy = (int)agent.getEnergy();
				
				viewport.drawString( agent.getId(), agent.getX() - 16, agent.getY() + 20, Color.BLACK );
				viewport.fillRect( agent.getX() - energy/4 - 1, agent.getY() - 21, energy/2 + 2, 5, Color.BLACK );
				viewport.fillRect( agent.getX() - energy/4, agent.getY() - 20, energy/2, 3, Color.YELLOW );
				
				if( agent.getSquad().contains( agents.get( lockedAgent ) ) ) {
					viewport.rect( agent.getX() - 25, agent.getY() - 25, 50, 50, Color.BLACK );
				}
			}
			
		}
		
		if( agents.get( lockedAgent ) != null ) {
			viewport.fillRect( 0, 0, this.viewport.getWidth(), 30, Color.BLACK, false );
			viewport.drawString( "Camera locked on: " + agents.get( lockedAgent ), 12, 12, Color.WHITE, false );
		}
	}

	public void startAgents() {
		for( Agent agent : agents ) {			
			Thread agentThread = new Thread( agent );
			agentThread.start();
		}
	}
	
	public void load( String definitionFile ) throws FileNotFoundException {
		Scanner scanner = new Scanner( new File( definitionFile ) );
		
		String mapName = scanner.nextLine();
		
		System.out.println( "Loading map: " + mapName );
		try {
			GameMap.getInstance().loadMap( mapName );
		} catch (FileNotFoundException e) {
			System.out.println("ERROR: Could not find map file!!!");
		}
		
		int maxSquadSize = scanner.nextInt();
		
		System.out.println("Maximum squad size: " + maxSquadSize);
		
		int agentEnergy = scanner.nextInt();
		
		System.out.println("Agent energy: " + agentEnergy);
		
		int numberOfAgents = scanner.nextInt();
		
		System.out.println("Number of agents: " + numberOfAgents);
		
		scanner.nextLine();
		while( numberOfAgents > 0 ) {
			StringTokenizer tk = new StringTokenizer( scanner.nextLine() );
			
			String id = tk.nextToken();
			String tId = tk.nextToken();
			int teamId = 0;
			
			if( tId.equals( "blue" ) ) {
				teamId = Agent.TEAM_BLUE;
			} else if( tId.equals( "red" ) ) {
				teamId = Agent.TEAM_RED;
			} else if( tId.equals( "green" ) ) {
				teamId = Agent.TEAM_GREEN;
			} else if( tId.equals( "yellow" ) ) {
				teamId = Agent.TEAM_YELLOW;
			}
			
			System.out.println("Adding agent: " + id + " on team " + tId  + " (" + teamId + ")");
			
			try {
				this.addAgent( id, teamId, agentEnergy, maxSquadSize );
			} catch (IOException e) {
				System.err.println( "Could not create agent!" );
			}
			
			numberOfAgents--;
		}
	}

	public boolean isGameOver() {
		return hasTeamWon( Agent.TEAM_BLUE ) ||
				hasTeamWon( Agent.TEAM_RED ) ||
				hasTeamWon( Agent.TEAM_GREEN ) ||
				hasTeamWon( Agent.TEAM_YELLOW );
	}
	
	public boolean hasTeamWon( int teamId ) {
		for( Agent agent : agents ) {
			if( agent.isAlive() && agent.getTeamId() != teamId ) {
				return false;
			}
		}
		
		return true;
	}
}
