package nom.paolo.ai.game.command;

import java.util.List;

import nom.paolo.ai.game.map.GameMap;
import nom.paolo.ai.game.objects.Agent;
import nom.paolo.ai.pathfinder.CartesianPathFinder;
import nom.paolo.ai.pathfinder.HeuristicPathFinder;
import nom.paolo.ai.pathfinder.PathComponent;

/**
 * Comand that moves the agent towards a point
 * 
 * @author Paolo Victor
 */
public class MoveAgentCommand extends AbstractAgentCommand {	
	private int destX;
	private int destY;
	
	private int tilesize;
	
	private List<PathComponent> path;

	private double speed;
	
	private int yOffset;
	private int xOffset;
	private boolean halted = false;
	
	public MoveAgentCommand(String name, Agent target) {
		super(name, target);
	}

		
	public MoveAgentCommand( Agent agent, int x, int y, double speed ) {
		this( "Move", agent );
		
		this.agent = agent;
		this.speed = speed;		
		this.tilesize = GameMap.getInstance().getTileSize();
		
		this.xOffset = x % tilesize;
		this.yOffset = y % tilesize;
		
		CartesianPathFinder pathfinder = new CartesianPathFinder( new HeuristicPathFinder( GameMap.getInstance().getVisibleMap( agent.getTeamId() ) ), tilesize );
		this.path = pathfinder.findPath( agent.getX(), agent.getY(), x, y );
				
		if( path != null ) {
			this.goToNextPathComponent();
		}
	}
	
	/**
	 * Retrieves the next path component from the component list
	 */
	private void goToNextPathComponent() {
		if( ! path.isEmpty() ) { 
			PathComponent next = path.remove(0);
		
			this.destX = next.getX() * tilesize + (tilesize/2) + xOffset/4;
			this.destY = next.getY() * tilesize + (tilesize/2) + yOffset/4;
		}
	}

	/**
	 * Executes the action
	 */
	public void execute() {				
		// Moving the agent
		int aX = agent.getX();
		int aY = agent.getY();
		
		if( ! (Math.abs( aX - destX ) <= speed) ) {
			if( aX < destX ) {
				agent.moveRight( speed );
			} else if( aX > destX ) {
				agent.moveLeft( speed );
			}
		} else {
			agent.moveRight( 0 );
		}
		
		if( ! (Math.abs( aY - destY ) <= speed) ) {
			if( aY > destY ) {
				agent.moveUp( speed );
			} else if( aY < destY ) {
				agent.moveDown( speed );
			}
		} else {
			agent.moveDown( 0 );
		}
		
		// Checking if the destination has been reached, going to next path component		
		if( reachedDestination() ) {
			this.goToNextPathComponent();
		}		

		// If it's finished, stop.
		if( this.isFinished() ) {
			agent.stopMoving();
			return;
		}
	}
	
	/**
	 * @return If the agent has reached the path destination
	 */
	public boolean reachedDestination() {		
		return ( Math.abs( agent.getX() - destX ) <= speed ) &&
			   ( Math.abs( agent.getY() - destY ) <= speed );
	}

	/**
	 * @return true if the action is finished
	 */
	public boolean isFinished() {
		return path == null || halted || ( path.isEmpty() && reachedDestination() ) ;  
	}

	/**
	 * Stops the action 
	 */
	public void stop() {
		this.agent.stopMoving();
		this.halted = true;
	}
}
