package nom.paolo.ai.pathfinder;
import java.util.LinkedList;
import java.util.List;

import nom.paolo.ai.util.MathUtils;

/**
 * Pathfinding algoritm.
 * 
 * @author Paolo Victor
 */
public class HeuristicPathFinder implements PathFinder {
	/**
	 * +---> y
	 * |
	 * v
	 * x
	 */
	private int[][] map;
	
	public List<PathComponent> selectedPath = null;
	
	public HeuristicPathFinder( int[][] map ) {
		this.map = map;
	}

	/**
	 * @return the path between two points, represented by a list
	 * of PathComponents
	 */
	public List<PathComponent> findPath( int startX, int startY, int targetX, int targetY ) {
		selectedPath = null;	
				
		PathComponent initial = new PathComponent( startX, startY );
		PathComponent target = new PathComponent( targetX, targetY );
				
		if( initial.equals( target ) ) {
			selectedPath = new LinkedList<PathComponent>();
			selectedPath.add( initial );
		} else {		
			List<PathComponent> result = new LinkedList<PathComponent>();
			List<PathComponent> visited = new LinkedList<PathComponent>();		
			
			findPath2( initial, target, result, visited, map );
		}
		
		return selectedPath;
	}
	
	public int[][] getMap() {
		return this.map;
	}
	
	/**
	 * Finds a path between two path components
	 * 
	 * @param initial
	 * @param target
	 * @param path
	 * @param visited
	 * @param map
	 */
	private void findPath2( PathComponent initial, PathComponent target, List<PathComponent> path, List<PathComponent> visited, int[][] map ) {		
		// If it's out of bounds, return
		if( map == null || initial.getX() < 0 || initial.getY() < 0 || initial.getY() >= map.length || initial.getX() >= map[0].length ) {
			return; 
		}
		
		// Adds node to visited nodes
		visited.add( initial );
		
		// If there's a wall, return 
		if( map[ initial.getY() ][ initial.getX() ] != 0 ) {
			return;
		}
		
		// Adds node to path
		path.add( initial );		
				
		// If it's the target, set as path
		if( initial.equals( target ) ) {
			selectedPath = new LinkedList<PathComponent>( path );
			return;
		}

		// Build neighbour visit sequence list - Insertion sort
		List<PathComponent> visitSequence = new LinkedList<PathComponent>();
		for( int h = -1; h <= 1; h++ ) {
			for( int v = -1; v <= 1; v++ ) {
				if( h == v || h == -v ) {
					continue;
				}
				
				PathComponent newInitial = new PathComponent( initial.getX() - h, initial.getY() - v );			
				
				// Inserts elements using insertion sort. For this particular problem, it's O(6)
				int insertPosition = 0;
				for( int p = 0; p < visitSequence.size(); p++ ) {
					PathComponent e = visitSequence.get(p);
						
					double distanceOld = MathUtils.distanceBetween( e.getX(), e.getY(), target.getX(), target.getY() );
					double distanceInit = MathUtils.distanceBetween( newInitial.getX(), newInitial.getY(), target.getX(), target.getY() );
									
					if( distanceInit > distanceOld ) {
						insertPosition++;
					} else {
						break;
					}
				}
				
				visitSequence.add( insertPosition, newInitial );
			}
		}
		
		// If the path was not found, go through the neighbours		
		for( int n = 0; n < visitSequence.size(); n++ ) {
			PathComponent next = visitSequence.get( n );
			if( selectedPath == null &&
				!path.contains( next ) &&
				!visited.contains( next ) ) {
				findPath2( next, target, path, visited, map );
			}
		}
		
		path.remove( initial );
	}
}
